# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
InvocationHandler is a function which takes the instance method being invoked
and its arguments. It processes a service method invocation on a proxy
instance and returns the result. The return value should match the arity of the
proxied method and have the same types. The first argument will always be the
FContext.

ServiceMiddleware is a function which takes an InvocationHandler and returns
a new InvocationHandler. It's used to implement interceptor logic around API
calls. This can be used, for example, to implement retry policies on service
calls, logging, telemetry, or authentication and authorization.
ServiceMiddleware can be applied to both RPC services and pub/sub scopes.

Middleware example:

def logging_middleware(next):

    def invocation_handler(method, args):
        service = '%s.%s' % (method.im_self.__module__,
                             method.im_class.__name__)
        print '==== CALLING %s.%s ====' % (service, method.im_func.func_name)
        ret = next(method, args)
        print '==== CALLED  %s.%s ====' % (service, method.im_func.func_name)
        return ret

    return invocation_handler

"""


class Method(object):
    """Method contains an InvocationHandler and a handle to the method it
    proxies. This should only be used by generated code.
    """

    def __init__(self, method, middleware):
        """Initialize a new Method which proxies the given handler method. This
        should only be called by generated code.

        Args:
            method: instance method
            middleware: list of ServiceMiddleware
        """

        self._handler = _compose_middleware(method, middleware)
        self._proxied_method = method

    def __call__(self, *args):
        """Invoke the Method and return its results. The should only be called
        by generated code.
        """

        return self._handler(self._proxied_method, *args)

    def add_middleware(self, middleware):
        """Add the given middleware to the Method.
        This should only be called before the server is started.

            Args:
             middleware: ServiceMiddleware
         """

        handler = self._handler
        if middleware:
            for m in middleware:
                handler = m(handler)
        self._handler = handler


def _compose_middleware(method, middleware):
    """Apply the given ServiceMiddleware to the provided method and return
    an InvocationHandler.

    Args:
        method: instance method
        middleware: list of ServiceMiddleware

    Returns:
        InvocationHandler
    """

    def base_handler(method, args):
        return method(*args)

    handler = base_handler
    if middleware:
        for m in middleware:
            handler = m(handler)
    return handler



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import sys
from thrift.protocol.TProtocolDecorator import TProtocolDecorator
from thrift.protocol.TCompactProtocol import CLEAR, TCompactProtocol

from frugal.context import FContext, _OPID_HEADER, _CID_HEADER, _get_next_op_id
from frugal.util.headers import _Headers

_V0 = 0


def _state_reset_decorator(func):
    """
    Decorator that resets the state of the TCompactProtocol as a hacky
    workaround for when an exception  occurs so the protocol can be reused,
    i.e. if "REQUEST_TOO_LARGE" error is thrown. This is only required for
    the compact protocol as other protocols don't track internal state as a
    sanity check.
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if not isinstance(self._wrapped_protocol, TCompactProtocol):
            return func(self, *args, **kwargs)

        try:
            return func(self, *args, **kwargs)
        except Exception:
            self._wrapped_protocol.state = CLEAR
            raise

    return wrapper


class FProtocol(TProtocolDecorator, object):
    """
    FProtocol is an extension of thrift TProtocol with the addition of headers
    """

    def __init__(self, wrapped_protocol):
        """
        Initialize FProtocol.

        Args:
            wrapped_protocol: wrapped thrift protocol extending TProtocolBase.
        """
        self._wrapped_protocol = wrapped_protocol
        super(FProtocol, self).__init__(self._wrapped_protocol)

    def get_transport(self):
        """
        Return the extended TProtocolBase's underlying tranpsort

        Returns:
            TTransportBase
        """
        return self.trans

    @_state_reset_decorator
    def write_request_headers(self, context):
        """
        Write the request headers to the underlying TTranpsort.
        """

        self._write_headers(context.get_request_headers())

    @_state_reset_decorator
    def write_response_headers(self, context):
        """
        Write the response headers to the underlying TTransport.
        """

        self._write_headers(context.get_response_headers())

    def _write_headers(self, headers):
        buff = _Headers._write_to_bytearray(headers)
        self.get_transport().write(buff)

    def read_request_headers(self):
        """
        Reads the request headers out of the underlying TTransportBase and
        return an FContext

        Returns:
            FContext
        """
        headers = _Headers._read(self.get_transport())

        context = FContext()

        for key, value in headers.items():
            context.set_request_header(key, value)

        op_id = headers[_OPID_HEADER]
        context._set_response_op_id(op_id)
        # Put a new opid in the request headers so this context an be
        # used/propagated on the receiver
        context.set_request_header(_OPID_HEADER, _get_next_op_id())

        cid = context.correlation_id
        if cid:
            context.set_response_header(_CID_HEADER, cid)
        return context

    def read_response_headers(self, context):
        """
        Read the response headers from the underlying transport
        and set them on a given FContext.

        Returns:
            FContext
        """
        headers = _Headers._read(self.get_transport())

        for key, value in headers.items():
            # Don't want to overwrite the opid header we set for a propagated
            # response
            if key == _OPID_HEADER:
                continue
            context.set_response_header(key, value)

    @_state_reset_decorator
    def writeMessageBegin(self, name, ttype, seqid):
        self._wrapped_protocol.writeMessageBegin(name, ttype, seqid)

    @_state_reset_decorator
    def writeMessageEnd(self):
        self._wrapped_protocol.writeMessageEnd()

    @_state_reset_decorator
    def writeStructBegin(self, name):
        self._wrapped_protocol.writeStructBegin(name)

    @_state_reset_decorator
    def writeStructEnd(self):
        self._wrapped_protocol.writeStructEnd()

    @_state_reset_decorator
    def writeFieldBegin(self, name, ttype, fid):
        self._wrapped_protocol.writeFieldBegin(name, ttype, fid)

    @_state_reset_decorator
    def writeFieldEnd(self):
        self._wrapped_protocol.writeFieldEnd()

    @_state_reset_decorator
    def writeFieldStop(self):
        self._wrapped_protocol.writeFieldStop()

    @_state_reset_decorator
    def writeMapBegin(self, ktype, vtype, size):
        self._wrapped_protocol.writeMapBegin(ktype, vtype, size)

    @_state_reset_decorator
    def writeMapEnd(self):
        self._wrapped_protocol.writeMapEnd()

    @_state_reset_decorator
    def writeListBegin(self, etype, size):
        self._wrapped_protocol.writeListBegin(etype, size)

    @_state_reset_decorator
    def writeListEnd(self):
        self._wrapped_protocol.writeListEnd()

    @_state_reset_decorator
    def writeSetBegin(self, etype, size):
        self._wrapped_protocol.writeSetBegin(etype, size)

    @_state_reset_decorator
    def writeSetEnd(self):
        self._wrapped_protocol.writeSetEnd()

    @_state_reset_decorator
    def writeBool(self, bool_val):
        self._wrapped_protocol.writeBool(bool_val)

    @_state_reset_decorator
    def writeByte(self, byte):
        self._wrapped_protocol.writeByte(byte)

    @_state_reset_decorator
    def writeI16(self, i16):
        self._wrapped_protocol.writeI16(i16)

    @_state_reset_decorator
    def writeI32(self, i32):
        self._wrapped_protocol.writeI32(i32)

    @_state_reset_decorator
    def writeI64(self, i64):
        self._wrapped_protocol.writeI64(i64)

    @_state_reset_decorator
    def writeDouble(self, dub):
        self._wrapped_protocol.writeDouble(dub)

    @_state_reset_decorator
    def writeString(self, value):
        """
        Write a string to the protocol, if python 2, encode to utf-8
        bytes from a unicode string.
        """
        if sys.version_info[0] == 2 and isinstance(value, unicode):  # noqa: F821,E501
            self._wrapped_protocol.writeString(value.encode('utf-8'))
        else:
            self._wrapped_protocol.writeString(value)

    @_state_reset_decorator
    def writeBinary(self, value):
        self._wrapped_protocol.writeBinary(value)

    def readString(self):
        """
        Read a string from the protocol, if python 2, decode from utf-8
        bytes to a unicode string.
        """
        if sys.version_info[0] == 2:
            return self._wrapped_protocol.readString().decode('utf-8')
        return self._wrapped_protocol.readString()



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import logging
import struct

from thrift.transport.TTransport import TMemoryBuffer

logger = logging.getLogger(__name__)


class _FHttpException(Exception):
    def __init__(self, code, message=None):
        super(_FHttpException, self).__init__(message)
        self.code = code


class _FHttpRequest(object):
    """
    _FHttpRequest stores data from an http request in a generic format.
    """
    def __init__(self, headers=None, body=b''):
        if headers is None:
            headers = {}
        self.headers = headers
        self.body = body


class _FHttpResponse(object):
    """
    _FHttpResponse returns data to be sent in an http response in a generic
    format.
    """
    def __init__(self, status_code=200, headers=None, body=b''):
        self._status_code = status_code
        if headers is None:
            headers = {}
        self._headers = headers
        self._body = body

    @property
    def status_code(self):
        """
        Status code of the response as an integer.
        """
        return self._status_code

    @property
    def headers(self):
        """
        dict of the HTTP response headers.
        """
        return self._headers

    @property
    def body(self):
        """
        The http response body.
        """
        return self._body


class _FHttpRequestHandler(object):
    """
    _FHttpRequestHandler provides functionality to process rpcs from http.

    Reading from/writing to the network of a library of choice is left up to
    the caller.
    """
    def __init__(self, processor, protocol_factory):
        """
        Initializes a _FHttpRequestHandler.

        Args:
            processor: The processor to use to handle requests.
            protocol_factory: A protocol factory to serialize/deserialize
                              frugal requests.
        """
        self._processor = processor
        self._protocol_factory = protocol_factory

    def _preprocess_http_request(self, request):
        """
        Performs some common preprocessing on an http request.

        Args:
            request: A _FHttpRequest object.
        Returns:
            otrans: The output transport, used for getting the output data
            iprot: The input protocol, given to a processor.
            oprot: The output protocol, given to a processor.
            response_limit: The maximum allowed size of the response,
                            zero if unlimited.
        """
        response_limit = int(request.headers.get('x-frugal-payload-limit', 0))
        payload = base64.b64decode(request.body)

        # Need 4 bytes for the frame size, at a minimum.
        if len(payload) < 4:
            logger.exception("invalid request size %s", len(payload))
            raise _FHttpException(400)

        # Ensure expected frame size equals actual size.
        size = struct.unpack('!I', payload[:4])[0]
        length = len(payload) - 4
        if size != length:
            raise _FHttpException(
                400, message='Mismatch between expected frame ' +
                'size ({}) and actual size ({})'.format(size, length))

        itrans = TMemoryBuffer(payload[4:])
        otrans = TMemoryBuffer()
        iprot = self._protocol_factory.get_protocol(itrans)
        oprot = self._protocol_factory.get_protocol(otrans)
        return otrans, iprot, oprot, response_limit

    def _postprocess_http_request(self, otrans, response_limit):
        """
        Performs some common postprocessing to produce an http response.

        Args:
            otrans: The output transport.
            response_limit: The maximum allowed size of the response,
                            zero if unlimited.
        Returns:
            A _FHttpResponse to write back to the client.
        """
        output_data = otrans.getvalue()
        if len(output_data) > response_limit > 0:
            logger.exception('response limit exceeded')
            return _FHttpResponse(status_code=413)

        frame_len = struct.pack('!I', len(output_data))
        frame = base64.b64encode(frame_len + output_data)

        headers = {
            'content-type': 'application/x-frugal',
            'content-transfer-encoding': 'base64',
        }
        return _FHttpResponse(headers=headers, body=frame)

    def _handle_processor_exception(self, ex):
        """
        Handles an unexpected exception from a processor.

        Args:
            ex: The exception.
        Returns:
            A _FHttpResponse.
        """
        return _FHttpResponse(status_code=500, body=ex.message)

    def handle_http_request(self, request):
        """
        Handles an http rpc. This must be implemented because of differences
        in processors (synchronous v tornado coroutine v asyncio etc.). It
        is recommended to use _preprocess_http_request and
        _postprocess_http_request to make this easier. If those function are
        used correctly, only calling the processor and handling processor
        errors should be necessary for implementors.

        Args:
            request: A _FHttpRequest object.
        Returns:
            a _FHttpResponse object.
        """
        raise NotImplementedError()


class _FSynchronousHttpRequestHandler(_FHttpRequestHandler):
    """
    An http request handler for synchronous processors.
    """

    def handle_http_request(self, request):
        """
        Handle a given http request

        Args:
            request - http request to handle
        Returns:
            _FHttpResponse
        """
        try:
            otrans, iprot, oprot, response_limit = \
                self._preprocess_http_request(request)
        except _FHttpException as e:
            return _FHttpResponse(status_code=e.code)

        try:
            self._processor.process(iprot, oprot)
        except Exception as e:
            return self._handle_processor_exception(e)

        return self._postprocess_http_request(otrans, response_limit)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging
from threading import Lock

from thrift.Thrift import TApplicationException
from thrift.Thrift import TMessageType
from thrift.Thrift import TType

from frugal.exceptions import TApplicationExceptionType

logger = logging.getLogger(__name__)


class FProcessorFunction(object):
    """
    FProcessorFunction is a generic object that exposes a single process
    call, which is used to handle a method invocation. FProcessorFunction
    should be implemented by the generated code.
    """

    def __init__(self, handler, lock):
        """
            Args:
                handler: frugal.middleware.Method
                lock: threading.Lock
        """
        self._handler = handler
        self._lock = lock

    def process(self, ctx, iprot, oprot):
        """
        Process the request from the input protocol and write the
        response to the output protocol.

        Args:
            iprot: input FProtocol
            oprot: output FProtocol
        """
        pass

    def add_middleware(self, middleware):
        """
        Add the given middleware to the FProcessorFunction
        This should only be called before the server is started.

            Args:
             middleware: ServiceMiddleware
         """

        self._handler.add_middleware(middleware)


class FProcessor(object):
    """
    FProcessor is a generic object which operates upon an input stream and
    writes to some output stream. FProcessor extends object.
    """

    def process(self, iprot, oprot):
        """
        Process the request from the input protocol and write the
        response to the output protocol.

        Args:
            iprot: input FProtocol
            oprot: output FProtocol
        """
        pass

    def add_middleware(self, service_middleware):
        """
        Adds the given ServiceMiddleware to the FProcessor. This should
        only called before the server is started.

        Args:
            service_middleware: ServiceMiddleware
        """
        pass

    def get_annotations_map(self):
        """
        Return a deepcopy of the annotations map.
        """
        pass


class FBaseProcessor(FProcessor):
    """
    FBaseProcessor is a base implementation of FProcessor. FProcessors
    should extend this and map FProcessorFunctions. This should only be used
    by generated code.
    """

    def __init__(self):
        """
        Create new instance of FBaseProcessor that will process requests.
        """
        self._processor_function_map = {}
        self._annotations_map = {}
        self._write_lock = Lock()

    def add_to_processor_map(self, key, proc):
        """
        Register the given FProcessorFunction.

        Args:
            key: processor function name
            proc: FProcessorFunction
        """
        self._processor_function_map[key] = proc

    def add_to_annotations_map(self, method_name, annotation):
        """
        Register the given annotation dictionary

        Args:
            method_name: method name
            annotation: annotation dictionary
        """
        self._annotations_map[method_name] = annotation

    def get_annotations_map(self):
        """
        Return a deepcopy of the annotations map
        """
        return copy.deepcopy(self._annotations_map)

    def get_write_lock(self):
        """
        Return the write lock.
        """
        return self._write_lock

    def process(self, iprot, oprot):
        """
        Process an input protocol and output protocol

        Args:
            iprot: input FProtocol
            oport: ouput FProtocol

        Raises:
            TException: if processing fails.
        """
        context = iprot.read_request_headers()
        name, _, _ = iprot.readMessageBegin()

        processor_function = self._processor_function_map.get(name)

        # If the function was in our dict, call process on it.
        if processor_function:
            try:
                processor_function.process(context, iprot, oprot)
            except Exception:
                # Don't raise an exception because the server should still send
                # a response to the client.
                logging.exception(
                    "frugal: exception occurred while processing request with "
                    "correlation id %s", context.correlation_id)
            return

        logging.warn("frugal: client invoked unknown method %s on request " +
                     "with correlation id %s", name, context.correlation_id)
        iprot.skip(TType.STRUCT)
        iprot.readMessageEnd()

        ex = TApplicationException(
            type=TApplicationExceptionType.UNKNOWN_METHOD,
            message="Unknown function: {0}".format(name))

        with self._write_lock:
            oprot.write_response_headers(context)
            oprot.writeMessageBegin(name, TMessageType.EXCEPTION, 0)
            ex.write(oprot)
            oprot.writeMessageEnd()
            oprot.trans.flush()

    def add_middleware(self, middleware):
        """
        Add the given middleware to the FProcessor.
        This should only be called before the server is started.

        Args:
            middleware: ServiceMiddleware
        """

        if middleware and not isinstance(middleware, list):
            middleware = [middleware]

        for proc in self._processor_function_map.values():
            proc.add_middleware(middleware)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import struct
from typing import List

from nats.aio.client import Client
from thrift.Thrift import TApplicationException
from thrift.transport.TTransport import TMemoryBuffer

from frugal import _NATS_MAX_MESSAGE_SIZE
from frugal.aio.processor import FProcessor
from frugal.protocol import FProtocolFactory
from frugal.server import FServer
from frugal.transport import TMemoryOutputBuffer

logger = logging.getLogger(__name__)


class FNatsServer(FServer):
    """FNatsServer is an FServer that uses nats as the underlying transport."""
    def __init__(
            self,
            nats_client: Client,
            subjects: List[str],
            processor: FProcessor,
            protocol_factory: FProtocolFactory,
            queue=''
    ):
        self._nats_client = nats_client
        self._subjects = [subjects] if isinstance(subjects, str) else subjects
        self._processor = processor
        self._protocol_factory = protocol_factory
        self._queue = queue
        self._sub_ids = []

    async def serve(self):
        """Subscribe to the server subject and queue."""
        self._sub_ids = []
        for subject in self._subjects:
            self._sub_ids.append(await self._nats_client.subscribe_async(
                subject,
                queue=self._queue,
                cb=self._on_message_callback,
            ))
        logger.info('Frugal server running...')

    async def stop(self):
        """Unsubscribe from the server subject."""
        for sid in self._sub_ids:
            await self._nats_client.unsubscribe(sid)

    async def _on_message_callback(self, message):
        """The function to be executed when a message is received."""
        if not message.reply:
            logger.warn('no reply present, discarding message')
            return

        frame_size = struct.unpack('!I', message.data[:4])[0]
        if frame_size > _NATS_MAX_MESSAGE_SIZE - 4:
            logger.warning('frame size too large, dropping message')
            return

        # process frame, first four bytes are the frame size
        iprot = self._protocol_factory.get_protocol(
            TMemoryBuffer(message.data[4:])
        )
        otrans = TMemoryOutputBuffer(_NATS_MAX_MESSAGE_SIZE)
        oprot = self._protocol_factory.get_protocol(otrans)

        try:
            await self._processor.process(iprot, oprot)
        except TApplicationException:
            # Continue so the exception is sent to the client
            pass
        except Exception:
            return

        if len(otrans) == 4:
            return

        await self._nats_client.publish(message.reply, otrans.getvalue())



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import async_timeout

from thrift.protocol.TProtocol import TProtocolException
from thrift.transport.TTransport import TMemoryBuffer
from thrift.transport.TTransport import TTransportException

from frugal.aio.transport import FTransportBase
from frugal.context import _OPID_HEADER
from frugal.exceptions import TTransportExceptionType
from frugal.context import FContext
from frugal.util.headers import _Headers


class FAsyncTransport(FTransportBase):
    """
     FAsyncTransport is an extension of FTransportBase that asynchronous
     frameworks can implement. Implementations need only implement flush to
     send request data and call handle_response when asynchronous responses
     are received.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._futures = {}
        self._futures_lock = asyncio.Lock()

    async def oneway(self, context: FContext, payload):
        self._preflight_request_check(payload)
        try:
            with async_timeout.timeout(context.timeout / 1000):
                await self.flush(payload)
        except asyncio.TimeoutError:
            raise TTransportException(
                type=TTransportExceptionType.TIMED_OUT,
                message='request timed out'
            ) from None

    async def request(self, context: FContext, payload):
        self._preflight_request_check(payload)
        op_id = str(context._get_op_id())
        future = asyncio.Future()
        async with self._futures_lock:
            if op_id in self._futures:
                raise TTransportException(
                    type=TTransportExceptionType.UNKNOWN,
                    message="request already in flight for context"
                )
            self._futures[op_id] = future

        try:
            with async_timeout.timeout(context.timeout / 1000):
                await self.flush(payload)
                return TMemoryBuffer(await future)
        except asyncio.TimeoutError:
            raise TTransportException(
                type=TTransportExceptionType.TIMED_OUT,
                message='request timed out'
            ) from None
        finally:
            async with self._futures_lock:
                del self._futures[op_id]

    async def flush(self, payload):
        """Flush the payload to the server."""
        raise NotImplementedError('You must override this')

    async def handle_response(self, frame):
        """
        Complete the future associated with the data frame.

        Args:
            frame: The response frame
        """
        if not frame:
            return
        headers = _Headers.decode_from_frame(frame)
        op_id = headers.get(_OPID_HEADER, None)

        if not op_id:
            raise TProtocolException(message="Frame missing op_id")

        async with self._futures_lock:
            future = self._futures.get(op_id, None)
            if not future:
                return

            future.set_result(frame)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from struct import pack_into
from struct import unpack_from

from thrift.protocol.TProtocol import TProtocolException

logger = logging.getLogger(__name__)

_V0 = 0
# Code for big endian unsigned char
_UCHAR = '!B'
# Code for big endian unsigned int
_UINT = '!I'
_UINT_LENGTH = 4


class _Headers(object):

    @staticmethod
    def _write_to_bytearray(headers):
        """
        Writes a given dictionary to a bytearray object and returns it
        TODO : Point to Protocol doc in the repo

        Args:
            headers: dict of frugal headers to write
        Returns:
            bytearray containing binary headers
        """
        utf8_headers = [(key.encode('utf8'), val.encode('utf8'))
                        for (key, val) in headers.items()]
        size = sum([8 + len(key) + len(val) for (key, val) in utf8_headers])
        buff = bytearray(size + 5)

        pack_into(_UCHAR, buff, 0, _V0)
        pack_into(_UINT, buff, 1, size)

        offset = 5

        for key, value in utf8_headers:
            key_len = len(key)
            pack_into(_UINT, buff, offset, key_len)
            offset += 4

            pack_into('>{0}s'.format(str(key_len)), buff, offset, key)
            offset += key_len

            val_len = len(value)
            pack_into(_UINT, buff, offset, val_len)
            offset += 4

            pack_into('>{0}s'.format(str(val_len)), buff, offset, value)
            offset += val_len

        return buff

    @staticmethod
    def _read(buff1):
        buff = buff1.read(1)
        version = unpack_from(_UCHAR, buff[:1])[0]

        if version is not _V0:
            ex = TProtocolException(
                TProtocolException.BAD_VERSION,
                "Wrong Frugal version. Found {0}, wanted {1}.".format(
                    version, _V0)
            )
            logger.exception(ex)
            raise ex

        buff = buff1.read(4)
        size = unpack_from(_UINT, buff[0:4])[0]

        buff = buff1.read(size)

        return _Headers._read_pairs(buff, 0, size)

    @staticmethod
    def decode_from_frame(frame):
        """
        Decode a frugal message from a frame.
        """
        if len(frame) < 5:
            ex = TProtocolException(TProtocolException.INVALID_DATA,
                                    "Invalid frame size: {}".format(len(frame))
                                    )
            logger.exception(ex)
            raise ex

        version = unpack_from(_UCHAR, frame[0:1])[0]

        if version is not _V0:
            ex = TProtocolException(
                TProtocolException.BAD_VERSION,
                "Wrong Frugal version. Found {0}, wanted {1}."
                .format(version, _V0)
            )
            logger.exception(ex)
            raise ex

        headers_size = unpack_from(_UINT, frame[1:5])[0]

        return _Headers._read_pairs(frame, 5, headers_size + 5)

    @staticmethod
    def _read_pairs(buff, start, end):
        parsed_headers = {}
        i = start
        while i < end:
            name_size = unpack_from(_UINT, buff[i:i + 4])[0]
            i += 4

            if i > end or i + name_size > end:
                ex = TProtocolException(TProtocolException.INVALID_DATA,
                                        "invalid protocol header name size: {}"
                                        .format(name_size))
                logger.exception(ex)
                raise ex

            name = unpack_from('>{0}s'.format(name_size),
                               buff[i:i + name_size])[0]
            i += name_size

            val_size = unpack_from(_UINT, buff[i: i + 4])[0]
            i += 4

            if i > end or i + val_size > end:
                ex = TProtocolException(
                    TProtocolException.INVALID_DATA,
                    "invalid protocol header value size: {}".format(val_size)
                )
                logger.exception(ex)
                raise ex

            val = unpack_from('>{0}s'.format(val_size),
                              buff[i:i + val_size])[0]
            i += val_size
            parsed_headers[name.decode('utf8')] = val.decode('utf8')

        return parsed_headers



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import httplib
import logging

from thrift.transport.TTransport import TMemoryBuffer
from thrift.transport.TTransport import TTransportException
from tornado import gen
from tornado.httpclient import AsyncHTTPClient
from tornado.httpclient import HTTPError
from tornado.httpclient import HTTPRequest

from frugal.exceptions import TTransportExceptionType
from frugal.tornado.transport.transport import FTransportBase

logger = logging.getLogger(__name__)


class FHttpTransport(FTransportBase):
    def __init__(self, url, request_capacity=0, response_capacity=0,
                 get_request_headers=None):
        """
        Create an HTTP transport.

        Args:
            url: The url to send requests to.
            request_capacity: The maximum size allowed to be written in a
                              request. Set to 0 for no size restrictions.
            response_capacity: The maximum size allowed to be read in a
                               response. Set to 0 for no size restrictions.
            get_request_headers: An optional function that accepts an FContext.
                                 Should return a dictionary of additional
                                 request headers to be appended to the request
        """
        super(FHttpTransport, self).__init__(
            request_size_limit=request_capacity)
        self._url = url
        self._http = AsyncHTTPClient()
        self._get_request_headers = get_request_headers

        # create headers
        self._headers = {
            'content-type': 'application/x-frugal',
            'content-transfer-encoding': 'base64',
            'accept': 'application/x-frugal',
        }
        if response_capacity > 0:
            self._headers['x-frugal-payload-limit'] = str(response_capacity)

        self._execute = None

    @gen.coroutine
    def is_open(self):
        """Always returns True"""
        # Tornado requires we raise a special exception to return a value.
        raise gen.Return(True)

    @gen.coroutine
    def open(self):
        """No-op"""
        pass

    @gen.coroutine
    def close(self):
        """no-op"""
        pass

    @gen.coroutine
    def oneway(self, context, payload):
        """
        Write the current buffer. This transport detects oneway requests via
        via the payload size on the server response. Therefore, just call
        through to request.
        """
        yield self.request(context, payload)

    @gen.coroutine
    def request(self, context, payload):
        """
        Write the current buffer and return the response.
        """
        # construct headers for request
        request_headers = {}
        if self._get_request_headers is not None:
            request_headers = self._get_request_headers(context)
        # apply the default headers so their values cannot be modified
        request_headers.update(self._headers)

        self._preflight_request_check(payload)
        encoded = base64.b64encode(payload)
        request = HTTPRequest(self._url,
                              method='POST',
                              body=encoded,
                              headers=request_headers,
                              request_timeout=context.timeout / 1000.0
                              )

        try:
            response = yield self._http.fetch(request)
        except HTTPError as e:
            if e.code == httplib.REQUEST_ENTITY_TOO_LARGE:
                raise TTransportException(
                    type=TTransportExceptionType.RESPONSE_TOO_LARGE,
                    message='response was too large')

            # Tornado HttpClient uses 599 as the HTTP code to indicate a
            # request timeout
            if e.code == 599:
                raise TTransportException(
                    type=TTransportExceptionType.TIMED_OUT,
                    message='request timed out')

            message = 'response errored with code {0} and body {1}'.format(
                e.code, e.message
            )
            raise TTransportException(
                type=TTransportExceptionType.UNKNOWN,
                message=message)

        decoded = base64.b64decode(response.body)

        if len(decoded) < 4:
            raise TTransportException(
                type=TTransportExceptionType.UNKNOWN,
                message='invalid frame size')

        if len(decoded) == 4:
            # One-way method, drop response
            return

        raise gen.Return(TMemoryBuffer(decoded[4:]))



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import struct

from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType
from thrift.transport.TTransport import TMemoryBuffer


class TMemoryOutputBuffer(TMemoryBuffer, object):
    """
    An implementation of TMemoryBuffer using a bounded memory buffer. Writes
    that cause the buffer to exceed its size throw an FMessageSizeException.
    This implementation handles framing data.
    """

    def __init__(self, limit, value=None):
        """
        Create an instance of FBoundedMemoryBuffer where size is the
        maximum writable length of the buffer.

           Args:
               limit: integer size limit of the buffer
               value: optional data value to initialize the buffer with.
        """
        super(TMemoryOutputBuffer, self).__init__(value)
        self._limit = limit

    def write(self, buf):
        """
        Bounded write to buffer
        """
        if len(self) + len(buf) > self._limit > 0:
            self._buffer = TMemoryBuffer()
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message="Buffer size reached {}".format(self._limit))
        self._buffer.write(buf)

    def getvalue(self):
        # TODO make more efficient?
        data = self._buffer.getvalue()
        return struct.pack('!I', len(data)) + data

    def read(self, sz):
        raise Exception("don't call this")

    def __len__(self):
        return len(self.getvalue())



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from base64 import b64encode
from io import BytesIO
from struct import pack_into

from thrift.transport.TTransport import TTransportException

from frugal.exceptions import TTransportExceptionType
from frugal.transport import TSynchronousTransport


class TBaseHttpTransport(TSynchronousTransport):
    """
    Base synchronous transport implemented with HTTP.
    """

    def __init__(self, url, request_capacity=0, response_capacity=0,
                 headers=None, get_headers=None):
        """
        Initialize a new FBaseHttpTransport.

        Args:
            url: url of the Frugal server.
            request_capacity: max size allowed to be written in a request. Set
                              0 for no restriction.
            response_capacity: max size allowed to be read in a response. Set
                               0 for no restriction.
            headers: dict containing static headers.
            get_headers: func which returns dynamic headers per request.
        """

        self._url = url
        self._wbuff = BytesIO()
        self._rbuff = BytesIO()
        self._request_capacity = request_capacity
        self._response_capacity = response_capacity
        self._custom_headers = headers
        self._get_headers = get_headers
        self._timeout = None

    def set_timeout(self, timeout):
        """
        Set the request timeout.

        Args:
            timeout: request timeout in milliseconds.
        """
        self._timeout = timeout

    def isOpen(self):
        return True

    def open(self):
        pass

    def close(self):
        pass

    def read(self, sz):
        return self._rbuff.read(sz)

    def write(self, buf):
        size = len(buf) + len(self._wbuff.getvalue())
        if size + 4 > self._request_capacity > 0:
            self._wbuff = BytesIO()
            raise TTransportException(
                type=TTransportExceptionType.REQUEST_TOO_LARGE,
                message='Message exceeds {0} bytes, was {1} bytes'.format(
                    self._request_capacity, size + 4))

        self._wbuff.write(buf)

    def _get_headers_and_body(self):
        """
        Return the request headers and body.

        Returns:
            headers dict and base64-encoded body string.
        """

        data = self._wbuff.getvalue()
        self._wbuff = BytesIO()
        frame_size = len(data)
        if frame_size == 0:
            return None, None

        # Prepend the frame size to the message.
        buff = bytearray(4)
        pack_into('!I', buff, 0, frame_size)

        body = b64encode(buff + data)

        headers = {
            'Content-Type': 'application/x-frugal',
            'Content-Length': str(len(body)),
            'Content-Transfer-Encoding': 'base64',
        }

        if self._response_capacity:
            headers['x-frugal-payload-limit'] = str(self._response_capacity)

        if self._get_headers:
            headers.update(self._get_headers())

        if self._custom_headers:
            for name, value in self._custom_headers.items():
                headers[name] = value

        if 'User-Agent' not in headers:
            headers['User-Agent'] = 'Python/TBaseHttpTransport'

        return headers, body



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thrift.Thrift import TApplicationException
from thrift.transport.TTransport import TTransportException


class TTransportExceptionType(object):
    """Exception types for TTransportExceptions"""
    UNKNOWN = TTransportException.UNKNOWN
    NOT_OPEN = TTransportException.NOT_OPEN
    ALREADY_OPEN = TTransportException.ALREADY_OPEN
    TIMED_OUT = TTransportException.TIMED_OUT
    END_OF_FILE = TTransportException.END_OF_FILE

    REQUEST_TOO_LARGE = 100
    RESPONSE_TOO_LARGE = 101


class TApplicationExceptionType(object):
    """Exception types for TApplicationExceptions"""
    UNKNOWN = TApplicationException.UNKNOWN
    UNKNOWN_METHOD = TApplicationException.UNKNOWN_METHOD
    INVALID_MESSAGE_TYPE = TApplicationException.INVALID_MESSAGE_TYPE
    WRONG_METHOD_NATE = TApplicationException.WRONG_METHOD_NAME
    BAD_SEQUENCE_ID = TApplicationException.BAD_SEQUENCE_ID
    MISSING_RESULT = TApplicationException.MISSING_RESULT
    INTERNAL_ERROR = TApplicationException.INTERNAL_ERROR
    PROTOCOL_ERROR = TApplicationException.PROTOCOL_ERROR
    INVALID_TRANSFORM = TApplicationException.INVALID_TRANSFORM
    INVALID_PROTOCOL = TApplicationException.INVALID_PROTOCOL
    UNSUPPORTED_CLIENT_TYPE = TApplicationException.UNSUPPORTED_CLIENT_TYPE

    RESPONSE_TOO_LARGE = 100



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from setuptools import setup, find_packages

from frugal.version import __version__

setup(
    name='frugal',
    version=__version__,
    description='Frugal Python Library',
    maintainer='Messaging Team',
    maintainer_email='messaging@workiva.com',
    url='http://github.com/Workiva/frugal',
    packages=find_packages(exclude=('frugal.tests', 'frugal.tests.*')),
    install_requires=[
        "six==1.10.0",
        "thrift==0.10.0",
        "requests==2.12.5",
    ],
    extras_require={
        'tornado': ["nats-client==0.5.0"],
        'asyncio': [
            "async-timeout==1.1.0",
            "asyncio-nats-client==0.6.0",
            # DO NOT CHANGE without verifying asyncio nats with tls works
            "multidict==3.1.3",
            "aiohttp==2.3.8",
        ],
        'gae': ["webapp2==2.5.2"],
    }
)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class FScopeProvider(object):
    """FScopeProviders produce FScopeTransports and FProtocols for use with
    Frugal Publishers and Subscribers. This also provides a shim for adding
    middleware to a publisher or subscriber.
    """

    def __init__(self, pub_transport_factory, sub_transport_factory,
                 protocol_factory, middleware=None):
        """Initialize FScopeProvider.

        Args:
            pub_transport_factory: FPublisherTransportFactory
            sub_transport_factory: FSubscriberTransportFactory
            protocol_factory: FProtocolFactory
            middleware: list of ServiceMiddleware
        """
        self._pub_transport_factory = pub_transport_factory
        self._sub_transport_factory = sub_transport_factory
        self._protocol_factory = protocol_factory
        self._middleware = middleware or []

    def new_publisher(self):
        """Return a tupled FScopeTransport and FProtocol.
        Returns:
            (FScopeTransport, FProtocolFactory)
        """
        transport = self._pub_transport_factory.get_transport()
        return transport, self._protocol_factory

    def new_subscriber(self):
        transport = self._sub_transport_factory.get_transport()
        return transport, self._protocol_factory

    def get_middleware(self):
        """Return the ServiceMiddleware stored on this FScopeProvider."""
        return list(self._middleware)


class FServiceProvider(object):
    """FServiceProvider is the service equivalent of FScopeProvider. It produces
     FTransports and FProtocols for use by RPC service clients. The main
     purpose of this is to provide a shim for adding middleware to a client.
     """

    def __init__(self, transport, protocol_factory, middleware=None):
        """Initialize FServiceProvider.

        Args:
            transport: FTransport
            protocol_factory: FProtocolFactory
            middleware: list of ServiceMiddleware
        """
        self._transport = transport
        self._protocol_factory = protocol_factory
        self._middleware = middleware or []

    def get_transport(self):
        """Get the FTransport from the provider."""
        return self._transport

    def get_protocol_factory(self):
        """Get the FProtocolFactory from the provider."""
        return self._protocol_factory

    def get_middleware(self):
        """Return the ServiceMiddleware stored on this FServiceProvider."""
        return list(self._middleware)



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64

from aiohttp import web
from thrift.Thrift import TApplicationException
from thrift.transport.TTransport import TMemoryBuffer

from frugal.aio.processor import FProcessor
from frugal.protocol import FProtocolFactory
from frugal.transport import TMemoryOutputBuffer


def new_http_handler(processor: FProcessor,
                     protocol_factory: FProtocolFactory):
    """
    Returns a function that can be used as a request handler in an aiohttp
    web server.

    Args:
        processor: The processor to use to handle requests.
        protocol_factory: A protocol factory to serialize/deserialize
                          frugal requests.
    """
    async def handler(request: web.Request):
        headers = {
            'content-type': 'application/x-frugal',
            'content-transfer-encoding': 'base64',
        }

        # check for response size limit
        response_limit = request.headers.get('x-frugal-payload-limit') or 0
        if response_limit:
            response_limit = int(response_limit)

        # decode payload and process
        payload = base64.b64decode(await request.content.read())
        iprot = protocol_factory.get_protocol(TMemoryBuffer(payload[4:]))
        # TODO could be better with this limit
        otrans = TMemoryOutputBuffer(0)
        oprot = protocol_factory.get_protocol(otrans)
        try:
            await processor.process(iprot, oprot)
        except TApplicationException:
            # Continue so the exception is sent to the client
            pass
        except Exception:
            return web.Response(status=400)

        # write back response
        output_data = otrans.getvalue()
        if len(output_data) > response_limit > 0:
            return web.Response(status=413)

        output_payload = base64.b64encode(output_data)

        return web.Response(body=output_payload, headers=headers)

    return handler



# Copyright 2017 Workiva
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
from copy import copy
from frugal import _IS_PY2

# Header containing correlation id.
_CID_HEADER = "_cid"

# Header containing op id (uint64 as string).
_OPID_HEADER = "_opid"

# Header containing request timeout (milliseconds as string).
_TIMEOUT_HEADER = "_timeout"

_DEFAULT_TIMEOUT = 5 * 1000

# Global incrementing op id.
_OP_ID = 0


class FContext(object):
    """
    FContext is the context for a Frugal message. Every RPC has an FContext,
    which can be used to set request headers, response headers, and the request
    timeout. The default timeout is five seconds. An FContext is also sent with
    every publish message which is then received by subscribers.

    In addition to headers, the FContext also contains a correlation ID which
    can be used for distributed tracing purposes. A random correlation ID is
    generated for each FContext if one is not provided.

    FContext also plays a key role in Frugal's multiplexing support. A unique,
    per-request operation ID is set on every FContext before a request is made.
    This operation ID is sent in the request and included in the response,
    which is then used to correlate a response to a request. The operation ID
    is an internal implementation detail and is not exposed to the user.

    An FContext should belong to a single request for the lifetime of that
    request. It can be reused once the request has completed, though they
    should generally not be reused. This class is _not_ thread-safe.
    """

    def __init__(self, correlation_id=None, timeout=_DEFAULT_TIMEOUT):
        """
        Initialize FContext.

        Args:
            correlation_id: string identifier for distributed tracing purposes.
            timeout: number of milliseconds before request times out.
        """
        self._request_headers = {}
        self._response_headers = {}

        if not correlation_id:
            correlation_id = self._generate_cid()

        self._request_headers[_CID_HEADER] = correlation_id
        self._request_headers[_TIMEOUT_HEADER] = str(timeout)

        # Take the current op id and increment the counter
        self._request_headers[_OPID_HEADER] = _get_next_op_id()

    @property
    def correlation_id(self):
        """
        Return the correlation id for the FContext. This is used for
        distributed tracing purposes.
        """
        return self._request_headers.get(_CID_HEADER)

    def _get_op_id(self):
        """
        Return an int operation id for the FContext.  This is a unique long
        per operation. This is protected as operation ids are an internal
        implementation detail.
        """

        return int(self._request_headers.get(_OPID_HEADER))

    def _set_op_id(self, op_id):
        self._request_headers[_OPID_HEADER] = str(op_id)

    def _set_response_op_id(self, op_id):
        self._response_headers[_OPID_HEADER] = op_id

    def get_request_headers(self):
        """
        Returns request headers for this FConext.
        """
        return copy(self._request_headers)

    def get_request_header(self, key):
        """
        Returns request header for the specified key from the request
        headers dict.
        """

        return self._request_headers.get(key)

    def set_request_header(self, key, value):
        """
        Set a string key value pair in the request headers dictionary.
        Return the same FContext to allow for call chaining. Changing the
        op ID or correlation ID is disallowed.

        Args:
            key: string key to set in request headers
            value: string value to set for the given key

        Returns:
            FContext

        Throws:
            TypeError: if user passes non-string for key or value.
        """
        self._check_string(key)
        self._check_string(value)

        self._request_headers[key] = value
        return self

    def get_response_headers(self):
        return copy(self._response_headers)

    def get_response_header(self, key):
        return self._response_headers.get(key)

    def set_response_header(self, key, value):
        """
        Set a string key value pair in the response headers dictionary.
        Return the same FContext to allow for call chaining. Changing the
        op ID or correlation ID is disallowed.

        Args:
            key: string key to set in response headers
            value: string value to set for the given key

        Returns:
            FContext

        Raises:
            TypeError: if user passes non-string for key or value.
        """
        self._check_string(key)
        self._check_string(value)

        self._response_headers[key] = value
        return self

    def get_timeout(self):
        """
        Get the timeout for the FContext.
        """
        return int(self._request_headers.get(_TIMEOUT_HEADER))

    def set_timeout(self, timeout):
        """
        Sets the timeout for the FContext.

        Args:
            timeout: number of seconds
        """
        self._request_headers[_TIMEOUT_HEADER] = str(timeout)

    @property
    def timeout(self):
        """
        Get the timeout for the FContext.
        """
        return int(self._request_headers.get(_TIMEOUT_HEADER))

    @timeout.setter
    def timeout(self, timeout):
        """
        Sets the timeout for the FContext.

        Args:
            timeout: number of seconds
        """
        # TODO: check the type of timeout
        self._request_headers[_TIMEOUT_HEADER] = str(timeout)
        return self

    def copy(self):
        """
        Performs a deep copy of an FContext while handling opids correctly.

        Returns:
            A new instance of FContext with identical headers, with the
            exception of _opid.
        """
        copied = FContext()
        copied._request_headers = self.get_request_headers()
        copied._response_headers = self.get_response_headers()
        copied._request_headers[_OPID_HEADER] = _get_next_op_id()
        return copied

    def _check_string(self, string):
        if _IS_PY2 and not \
                (isinstance(string, str) or isinstance(string, unicode)):  # noqa: F821,E501
            raise TypeError("Value should either be a string or unicode.")
        if not _IS_PY2 and not \
                (isinstance(string, str) or isinstance(string, bytes)):
            raise TypeError("Value should be either a string or bytes.")

    def _generate_cid(self):
        return uuid.uuid4().hex


def _get_next_op_id():
    global _OP_ID
    _OP_ID += 1
    return str(_OP_ID)
